/*----------------------------------------------------------------------
  File    : chull.h
  Contents: determine a convex hull in 3-space
  Author  : Christian Borgelt
  History : 19.10.2000 file created
            21.10.2000 fields px, py and edge added to structure POINT
            09.12.2000 parameter 'visib' added to function ch_draw()
            07.08.2001 POINT changed to POINT3 (conflict with windows.h)
----------------------------------------------------------------------*/
#ifndef __CHULL__
#define __CHULL__
#include <stdio.h>
#ifdef CH_DRAW
#include "proj3d.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- types of random points --- */
#define CH_CUBE          0      /* point in a cube [-1,1]^3 */
#define CH_SPHERE        1      /* point in a sphere in [-1,1]^3 */
#define CH_SURFACE       2      /* point on the surface of a sphere */

/* --- algorithm identifiers --- */
#define CH_SORTED        0      /* sorted points algorithm */
#define CH_CGRAPH        1      /* conflict graph algorithm */

#ifdef CH_DRAW
/* --- drawing modes --- */
#define CH_LINE       0x01      /* draw lines to the origin */
#define CH_MARK       0x02      /* mark points */
#define CH_WFRAME        0      /* draw as a wireframe model */
#define CH_SOLID         1      /* draw as a solid body */
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct cfle {           /* --- conflict list element --- */
  struct cfle  *pt_succ;        /* successor   in a point's list */
  struct cfle  *pt_pred;        /* predecessor in a point's list */
  struct cfle  *fc_succ;        /* successor   in a facet's list */
  struct cfle  *fc_pred;        /* predecessor in a facet's list */
  struct point *point;          /* point and */
  struct facet *facet;          /* facet that are in conflict */
} CFLE;                         /* (conflict list element) */

typedef struct point {          /* --- a point in 3-space --- */
  struct point *succ;           /* successor   in point list */
  struct point *pred;           /* predecessor in point list */
  double       x, y, z;         /* coordinates */
  CFLE         *confls;         /* list of conflicts */
  struct edge  *edge;           /* for cyclic list of horizon edges */
  int          mark;            /* marker for internal use */
  int          refcnt;          /* reference counter */
  #ifdef CH_DRAW                /* if drawing function is requested */
  double       px, py;          /* coordinates of projection */
  #endif                        /* to avoid recomputations */
} POINT3;                       /* (a point in 3-space) */

typedef struct edge {           /* --- an edge in 3-space --- */
  struct edge  *succ;           /* successor   in edge list */
  struct edge  *pred;           /* predecessor in edge list */
  double       dx, dy, dz;      /* (unit) direction vector */
  POINT3       *points[2];      /* end points */
  struct facet *facets[2];      /* incident facets */
  int          pos[2];          /* positions in facets' edge vectors */
  int          refcnt;          /* reference counter */
  int          mark;            /* marker for drawing */
} EDGE;                         /* (an edge in 3-space) */

typedef struct facet {          /* --- a facet in 3-space --- */
  struct facet *succ;           /* successor   in facet list */
  struct facet *pred;           /* predecessor in facet list */
  double       nx, ny, nz;      /* (unit) of normal vector */
  int          edgevsz;         /* current size of edge vector */
  int          edgecnt;         /* number of incident edges */
  EDGE         **edges;         /* vector of incident edges */
  CFLE         *confls;         /* list of conflicts */
  int          visible;         /* whether facet is visible */
} FACET;                        /* (a facet in 3-space) */

typedef struct chull {          /* --- a convex hull in 3-space --- */
  int           ptcnt;          /* number of corner points */
  POINT3        *points;        /* list of points */
  int           edgecnt;        /* number of edges */
  EDGE          *edges;         /* list of edges */
  int           facetcnt;       /* number of facets */
  FACET         *facets;        /* list of facets */
  int           alg;            /* algorithm used for construction */
  int           dim;            /* current dimensionality */
  POINT3        *next;          /* next unprocessed point */
  EDGE          *last;          /* last edge in edge list */
  FACET         *visible;       /* facet visible from next point */
  int           init;           /* initial number of points */
  int           created;        /* number of created facets */
} CHULL;                        /* (a convex hull in 3-space) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern CHULL* ch_create (void);
extern void   ch_delete (CHULL *chull);

extern void   ch_rand   (int type, double randfn(void),
                         double *x, double *y, double *z);
extern int    ch_point  (CHULL *chull, double x, double y, double z);
extern int    ch_start  (CHULL *chull, int alg);
extern int    ch_next   (CHULL *chull);
extern int    ch_build  (CHULL *chull, int alg);

extern int    ch_save   (CHULL *chull, const char *fname);
extern int    ch_load   (CHULL *chull, const char *fname);

extern void   ch_stats  (CHULL *chull);

#ifdef CH_DRAW
extern void   ch_draw   (CHULL *chull, int visib, PROJ3D *p3d,
                         void drawfn (double x, double y, int op));
#endif
#endif
